# -*- coding: utf-8 -*-
# Copyright: Damien Elmes <anki@ichi2.net>
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

import  cgi
from anki.consts import NEW_CARDS_RANDOM
from anki.lang import _
from anki.utils import fieldChecksum, guid64, timestampID, \
    joinFields, intTime, splitFields
from anki.importing.base import Importer
from anki.lang import ngettext

# Stores a list of fields, tags and deck
######################################################################

class ForeignNote(object):
    "An temporary object storing fields and attributes."
    def __init__(self):
        self.fields = []
        self.tags = []
        self.deck = None
        self.cards = {} # map of ord -> card

class ForeignCard(object):
    def __init__(self):
        self.due = 0
        self.ivl = 1
        self.factor = 2500
        self.reps = 0
        self.lapses = 0

# Base class for CSV and similar text-based imports
######################################################################

# The mapping is list of input fields, like:
# ['Expression', 'Reading', '_tags', None]
# - None means that the input should be discarded
# - _tags maps to note tags
# If the first field of the model is not in the map, the map is invalid.

# The import mode is one of:
# 0: update if first field matches existing note
# 1: ignore if first field matches existing note
# 2: import even if first field matches existing note

class NoteImporter(Importer):

    needMapper = True
    needDelimiter = False
    allowHTML = False
    importMode = 0

    def __init__(self, col, file):
        Importer.__init__(self, col, file)
        self.model = col.models.current()
        self.mapping = None
        self._deckMap = {}

    def run(self):
        "Import."
        assert self.mapping
        c = self.foreignNotes()
        self.importNotes(c)

    def fields(self):
        "The number of fields."
        return 0

    def initMapping(self):
        flds = [f['name'] for f in self.model['flds']]
        # truncate to provided count
        flds = flds[0:self.fields()]
        # if there's room left, add tags
        if self.fields() > len(flds):
            flds.append("_tags")
        # and if there's still room left, pad
        flds = flds + [None] * (self.fields() - len(flds))
        self.mapping = flds

    def mappingOk(self):
        return self.model['flds'][0]['name'] in self.mapping

    def foreignNotes(self):
        "Return a list of foreign notes for importing."
        assert 0

    def open(self):
        "Open file and ensure it's in the right format."
        return

    def importNotes(self, notes):
        "Convert each card into a note, apply attributes and add to col."
        assert self.mappingOk()
        # note whether tags are mapped
        self._tagsMapped = False
        for f in self.mapping:
            if f == "_tags":
                self._tagsMapped = True
        # gather checks for duplicate comparison
        csums = {}
        for csum, id in self.col.db.execute(
            "select csum, id from notes where mid = ?", self.model['id']):
            if csum in csums:
                csums[csum].append(id)
            else:
                csums[csum] = [id]
        firsts = {}
        fld0idx = self.mapping.index(self.model['flds'][0]['name'])
        self._fmap = self.col.models.fieldMap(self.model)
        self._nextID = timestampID(self.col.db, "notes")
        # loop through the notes
        updates = []
        updateLog = []
        updateLogTxt = _("First field matched: %s")
        dupeLogTxt = _("Added duplicate with first field: %s")
        new = []
        self._ids = []
        self._cards = []
        self._emptyNotes = False
        dupeCount = 0
        dupes = []
        for n in notes:
            for c in range(len(n.fields)):
                if not self.allowHTML:
                    n.fields[c] = cgi.escape(n.fields[c])
                n.fields[c] = n.fields[c].strip()
                if not self.allowHTML:
                    n.fields[c] = n.fields[c].replace("\n", "<br>")
            fld0 = n.fields[fld0idx]
            csum = fieldChecksum(fld0)
            # first field must exist
            if not fld0:
                self.log.append(_("Empty first field: %s") %
                                " ".join(n.fields))
                continue
            # earlier in import?
            if fld0 in firsts and self.importMode != 2:
                # duplicates in source file; log and ignore
                self.log.append(_("Appeared twice in file: %s") %
                                fld0)
                continue
            firsts[fld0] = True
            # already exists?
            found = False
            if csum in csums:
                # csum is not a guarantee; have to check
                for id in csums[csum]:
                    flds = self.col.db.scalar(
                        "select flds from notes where id = ?", id)
                    sflds = splitFields(flds)
                    if fld0 == sflds[0]:
                        # duplicate
                        found = True
                        if self.importMode == 0:
                            data = self.updateData(n, id, sflds)
                            if data:
                                updates.append(data)
                                updateLog.append(updateLogTxt % fld0)
                                dupeCount += 1
                                found = True
                        elif self.importMode == 1:
                            dupeCount += 1
                        elif self.importMode == 2:
                            # allow duplicates in this case
                            if fld0 not in dupes:
                                # only show message once, no matter how many
                                # duplicates are in the collection already
                                updateLog.append(dupeLogTxt % fld0)
                                dupes.append(fld0)
                            found = False
            # newly add
            if not found:
                data = self.newData(n)
                if data:
                    new.append(data)
                    # note that we've seen this note once already
                    firsts[fld0] = True
        self.addNew(new)
        self.addUpdates(updates)
        # make sure to update sflds, etc
        self.col.updateFieldCache(self._ids)
        # generate cards
        if self.col.genCards(self._ids):
            self.log.insert(0, _(
                "Empty cards found. Please run Tools>Empty Cards."))
        # apply scheduling updates
        self.updateCards()
        # we randomize or order here, to ensure that siblings
        # have the same due#
        did = self.col.decks.selected()
        conf = self.col.decks.confForDid(did)
        # in order due?
        if conf['new']['order'] == NEW_CARDS_RANDOM:
            self.col.sched.randomizeCards(did)
        else:
            self.col.sched.orderCards(did)

        part1 = ngettext("%d note added", "%d notes added", len(new)) % len(new)
        part2 = ngettext("%d note updated", "%d notes updated",
                         self.updateCount) % self.updateCount
        if self.importMode == 0:
            unchanged = dupeCount - self.updateCount
        elif self.importMode == 1:
            unchanged = dupeCount
        else:
            unchanged = 0
        part3 = ngettext("%d note unchanged", "%d notes unchanged",
                         unchanged) % unchanged
        self.log.append("%s, %s, %s." % (part1, part2, part3))
        self.log.extend(updateLog)
        if self._emptyNotes:
            self.log.append(_("""\
One or more notes were not imported, because they didn't generate any cards. \
This can happen when you have empty fields or when you have not mapped the \
content in the text file to the correct fields."""))
        self.total = len(self._ids)

    def newData(self, n):
        id = self._nextID
        self._nextID += 1
        self._ids.append(id)
        if not self.processFields(n):
            return
        # note id for card updates later
        for ord, c in n.cards.items():
            self._cards.append((id, ord, c))
        self.col.tags.register(n.tags)
        return [id, guid64(), self.model['id'],
                intTime(), self.col.usn(), self.col.tags.join(n.tags),
                n.fieldsStr, "", "", 0, ""]

    def addNew(self, rows):
        self.col.db.executemany(
            "insert or replace into notes values (?,?,?,?,?,?,?,?,?,?,?)",
            rows)

    def updateData(self, n, id, sflds):
        self._ids.append(id)
        if not self.processFields(n, sflds):
            return
        if self._tagsMapped:
            self.col.tags.register(n.tags)
            tags = self.col.tags.join(n.tags)
            return [intTime(), self.col.usn(), n.fieldsStr, tags,
                    id, n.fieldsStr, tags]
        else:
            return [intTime(), self.col.usn(), n.fieldsStr,
                    id, n.fieldsStr]

    def addUpdates(self, rows):
        old = self.col.db.totalChanges()
        if self._tagsMapped:
            self.col.db.executemany("""
update notes set mod = ?, usn = ?, flds = ?, tags = ?
where id = ? and (flds != ? or tags != ?)""", rows)
        else:
            self.col.db.executemany("""
update notes set mod = ?, usn = ?, flds = ?
where id = ? and flds != ?""", rows)
        self.updateCount = self.col.db.totalChanges() - old

    def processFields(self, note, fields=None):
        if not fields:
            fields = [""]*len(self.model['flds'])
        for c, f in enumerate(self.mapping):
            if not f:
                continue
            elif f == "_tags":
                note.tags.extend(self.col.tags.split(note.fields[c]))
            else:
                sidx = self._fmap[f][0]
                fields[sidx] = note.fields[c]
        note.fieldsStr = joinFields(fields)
        ords = self.col.models.availOrds(self.model, note.fieldsStr)
        if not ords:
            self._emptyNotes = True
        return ords

    def updateCards(self):
        data = []
        for nid, ord, c in self._cards:
            data.append((c.ivl, c.due, c.factor, c.reps, c.lapses, nid, ord))
        # we assume any updated cards are reviews
        self.col.db.executemany("""
update cards set type = 2, queue = 2, ivl = ?, due = ?,
factor = ?, reps = ?, lapses = ? where nid = ? and ord = ?""", data)



#! /usr/bin/python

'''
Anki card creator.
'''
import os, pdb, csv, random, re, sys
from anki import Collection
from anki.importing import TextImporter
from abc import ABCMeta, abstractmethod, abstractproperty

class HTML(object):

    '''
    Helper methods for creating HTML styling.
    '''
    @staticmethod
    def add_div_tag(text, class_name=None, id_name=None):
        return HTML.add_tag(text, 'div', class_name, id_name)

    @staticmethod
    def add_span_tag(text, class_name=None, id_name=None):
        return HTML.add_tag(text, 'span', class_name, id_name)

    @staticmethod
    def add_tag(text, tag_name, class_name=None, id_name=None):
        class_text = u''
        id_text = u''
        if class_name:
            class_text = u'class="{}"'.format(class_name)
        if id_name:
            id_text = u'id="{}"'.format(id_name)
        return u'<{} {} {}> {} </{}>'.format(tag_name,
                                            class_text,
                                            id_text,
                                            text,
                                            tag_name)
    @staticmethod
    def substitute_clozure(text, word):
        text = re.sub(word, u'{{{{c1::{}}}}}'.format(word), text, flags=re.I)
        return text


class AnkiObject(object):

    '''
    Stores information used to create an anki card.
    '''
    __metaclass__  = ABCMeta

    @abstractproperty
    def tag_name(self):
        pass

    @abstractproperty
    def model_name(self):
        pass

    def __init__(self, answer):
        self.answer = answer

    @abstractmethod
    def question_text(self):
        pass

    @abstractmethod
    def styling_text(self):
        pass

def get_delimiter():
    return '\t'

def get_question_index():
    return 1

def get_answer_index():
    return 0

def get_card_ids(collection, tag_name):
    ids = collection.findCards('tag:{}'.format(tag_name))
    return ids

def get_answers_from_collection(collection, ids):
    answers = []
    for id in ids:
        card = collection.getCard(id)
        note = card.note()
        answer = note.fields[get_answer_index()]
        answers.append(answer)
    return answers

def get_all_answers_from_collection(collection, tag_name):
    ids = get_card_ids(collection, tag_name)
    return get_answers_from_collection(collection, ids)

def sample_answers_from_collection(collection, tag_name, num_samples):
    ids = get_card_ids(collection, tag_name)
    error_msg = "Need at least {} in collection.".format(num_samples)
    assert len(ids) >= num_samples, error_msg
    sampled_ids = random.sample(ids, num_samples)

    return get_answers_from_collection(collection, sampled_ids)

def sample_answers_from_objects(anki_objects, num_samples):

    assert len(anki_objects) >= num_samples, "Not enough samples."

    sample_objects = random.sample(anki_objects,
                                         num_samples)
    samples = map(lambda anki_object : anki_object.answer,
                  sample_objects)
    return samples

def create_csv_row(anki_object, anki_objects, tag_name, collection,
                   num_choices=4):

    question = anki_object.question_text()
    answer = anki_object.answer

    other_objects = list(anki_objects)
    other_objects.remove(anki_object)

    num_from_objects = min(num_choices - 1, len(other_objects))
    object_samples = sample_answers_from_objects(other_objects,
                                                 num_from_objects)

    num_from_collection = num_choices - 1 - num_from_objects
    collection_samples = sample_answers_from_collection(collection,
                                                 tag_name,
                                                 num_from_collection)

    choices = object_samples + collection_samples
    choices += [anki_object.answer]
    random.shuffle(choices)
    letters = u'abcde'
    choices_text = ['{}) {}'.format(letter, choice)
                for (letter, choice) in zip(letters, choices)]
    choices_text = [HTML.add_div_tag(choice, class_name="choice")
                for choice in choices_text]

    choices_text = u''.join(choices_text)
    answer_text = choices_text
    answer_text = re.sub(answer,
                         u'<b>{}</b>'.format(answer),
                         choices_text)

    row = [answer, question, choices_text, answer_text]
    return row

def write_to_csv(anki_objects, tag_name, file_path, collection,
                 num_choices):

    with open(file_path, 'wb') as csvfile:
        writer = csv.writer(csvfile, delimiter=get_delimiter())

        for anki_object in anki_objects:
            answer = anki_object.answer
            if answer in get_all_answers_from_collection(collection,
                                                         tag_name):
                print '"{}" already in collection. Skipping.'.format(answer)
                continue
            row = create_csv_row(anki_object, anki_objects,
                                 tag_name, collection, num_choices)
            # convert back to bytes
            row = [cell.encode('utf-8') for cell in row]
            writer.writerow(row)

def set_model(model_name, deck_name, collection):
    did = collection.decks.id(deck_name)
    collection.decks.select(did)
    model = collection.models.byName(model_name)
    deck = collection.decks.get(did)
    deck['mid'] = model['id']
    collection.decks.save(deck)

def create_fields(collection):
    models = collection.models
    field_names = ['Value', 'Question', 'Choices', 'Answers']
    fields = map(lambda field_name:models.newField(field_name),
                 field_names)
    return fields

def create_templates(collection):
    models = collection.models
    template = models.newTemplate(u'Card 1')

    # SOMEDAY: Make fields not hard-coded
    question_text = '{{cloze:Question}}{{Choices}}'
    answer_text = '{{cloze:Question}}{{Answers}}'
    template['qfmt'] = question_text
    template['afmt'] = answer_text
    return [template]

def create_model(anki_object, deck_name, collection):
    model_name = anki_object.model_name
    model = collection.models.byName(model_name)
    if model is None:
        print 'Creating model: {}'.format(model_name)

        models = collection.models
        model = models.new(model_name)
        models.add(model)

        for field in create_fields(collection):
            models.addField(model, field)

        for template in create_templates(collection):
            models.addTemplate(model, template)

        model['css'] = anki_object.styling_text

        collection.models.update(model)

    set_model(model_name, deck_name, collection)


def run_importer(file_path, tag_name, deck_name, collection):
    if not os.stat(file_path).st_size > 0:
        print "Nothing to import!"
        sys.exit()
    ti = TextImporter(collection, file_path)
    ti.delimiter = get_delimiter()
    ti.allowHTML = True
    ti.tagsToAdd = [tag_name]
    ti.initMapping()
    ti.run()

    # BUGFIX: anki doesn't add to selected deck
    did = collection.decks.id(deck_name)
    num_cards_added = ti.total
    ids = get_card_ids(collection, tag_name)
    ids = sorted(ids, reverse=True)
    for id in ids[:num_cards_added]:
        collection.db.execute("update cards set did = ? where id = ?",
                              did, id)


def import_to_anki(file_path, tag_name, deck_name, collection):

    assert deck_name in collection.decks.allNames(), "No deck named " + deck_name

    run_importer(file_path, tag_name, deck_name, collection)

def get_csv_file_path():
    file_name = 'temp.csv'
    dir_name = os.path.dirname(os.path.realpath(__file__))
    return os.path.join(dir_name, file_name)

def get_collection(collection_path):
    assert os.path.exists(collection_path), "No collection found."
    # BUGFIX: Opening collection changes the cwd.
    cwd = os.getcwd()
    collection = Collection(collection_path)
    os.chdir(cwd)
    return collection

def create_cards(anki_objects, deck_name, collection_path, num_choices):

    collection = get_collection(collection_path)
    csv_file_path = get_csv_file_path()
    tag_name = anki_objects[0].tag_name

    write_to_csv(anki_objects, tag_name, csv_file_path, collection,
                 num_choices)
    create_model(anki_objects[0], deck_name, collection)
    import_to_anki(csv_file_path, tag_name, deck_name, collection)

    collection.close()




# -*- coding: utf-8 -*-
# Copyright: Damien Elmes <anki@ichi2.net>
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

import os
import unicodedata
from anki import Collection
from anki.utils import intTime, splitFields, joinFields, incGuid
from anki.importing.base import Importer
from anki.lang import _
from anki.lang import ngettext

GUID = 1
MID = 2
MOD = 3

class Anki2Importer(Importer):

    needMapper = False
    deckPrefix = None
    allowUpdate = True
    dupeOnSchemaChange = False

    def run(self, media=None):
        self._prepareFiles()
        if media is not None:
            # Anki1 importer has provided us with a custom media folder
            self.src.media._dir = media
        try:
            self._import()
        finally:
            self.src.close(save=False)

    def _prepareFiles(self):
        self.dst = self.col
        self.src = Collection(self.file)

    def _import(self):
        self._decks = {}
        if self.deckPrefix:
            id = self.dst.decks.id(self.deckPrefix)
            self.dst.decks.select(id)
        self._prepareTS()
        self._prepareModels()
        self._importNotes()
        self._importCards()
        self._importStaticMedia()
        self._postImport()
        self.dst.db.execute("vacuum")
        self.dst.db.execute("analyze")

    # Notes
    ######################################################################

    def _importNotes(self):
        # build guid -> (id,mod,mid) hash & map of existing note ids
        self._notes = {}
        existing = {}
        for id, guid, mod, mid in self.dst.db.execute(
            "select id, guid, mod, mid from notes"):
            self._notes[guid] = (id, mod, mid)
            existing[id] = True
        # we may need to rewrite the guid if the model schemas don't match,
        # so we need to keep track of the changes for the card import stage
        self._changedGuids = {}
        # apart from upgrading from anki1 decks, we ignore updates to changed
        # schemas. we need to note the ignored guids, so we avoid importing
        # invalid cards
        self._ignoredGuids = {}
        # iterate over source collection
        add = []
        update = []
        dirty = []
        usn = self.dst.usn()
        dupes = 0
        dupesIgnored = []
        for note in self.src.db.execute(
            "select * from notes"):
            # turn the db result into a mutable list
            note = list(note)
            shouldAdd = self._uniquifyNote(note)
            if shouldAdd:
                # ensure id is unique
                while note[0] in existing:
                    note[0] += 999
                existing[note[0]] = True
                # bump usn
                note[4] = usn
                # update media references in case of dupes
                note[6] = self._mungeMedia(note[MID], note[6])
                add.append(note)
                dirty.append(note[0])
                # note we have the added the guid
                self._notes[note[GUID]] = (note[0], note[3], note[MID])
            else:
                # a duplicate or changed schema - safe to update?
                dupes += 1
                if self.allowUpdate:
                    oldNid, oldMod, oldMid = self._notes[note[GUID]]
                    # will update if incoming note more recent
                    if oldMod < note[MOD]:
                        # safe if note types identical
                        if oldMid == note[MID]:
                            # incoming note should use existing id
                            note[0] = oldNid
                            note[4] = usn
                            note[6] = self._mungeMedia(note[MID], note[6])
                            update.append(note)
                            dirty.append(note[0])
                        else:
                            dupesIgnored.append("%s: %s" % (
                                self.col.models.get(oldMid)['name'],
                                note[6].replace("\x1f", ",")
                            ))
                            self._ignoredGuids[note[GUID]] = True
        if dupes:
            up = len(update)
            self.log.append(_("Updated %(a)d of %(b)d existing notes.") % dict(
                a=len(update), b=dupes))
            if dupesIgnored:
                self.log.append(_("Some updates were ignored because note type has changed:"))
                self.log.extend(dupesIgnored)
        # export info for calling code
        self.dupes = dupes
        self.added = len(add)
        self.updated = len(update)
        # add to col
        self.dst.db.executemany(
            "insert or replace into notes values (?,?,?,?,?,?,?,?,?,?,?)",
            add)
        self.dst.db.executemany(
            "insert or replace into notes values (?,?,?,?,?,?,?,?,?,?,?)",
            update)
        self.dst.updateFieldCache(dirty)
        self.dst.tags.registerNotes(dirty)

    # determine if note is a duplicate, and adjust mid and/or guid as required
    # returns true if note should be added
    def _uniquifyNote(self, note):
        origGuid = note[GUID]
        srcMid = note[MID]
        dstMid = self._mid(srcMid)
        # duplicate schemas?
        if srcMid == dstMid:
            return origGuid not in self._notes
        # differing schemas and note doesn't exist?
        note[MID] = dstMid
        if origGuid not in self._notes:
            return True
        # as the schemas differ and we already have a note with a different
        # note type, this note needs a new guid
        if not self.dupeOnSchemaChange:
            return False
        while True:
            note[GUID] = incGuid(note[GUID])
            self._changedGuids[origGuid] = note[GUID]
            # if we don't have an existing guid, we can add
            if note[GUID] not in self._notes:
                return True
            # if the existing guid shares the same mid, we can reuse
            if dstMid == self._notes[note[GUID]][MID]:
                return False

    # Models
    ######################################################################
    # Models in the two decks may share an ID but not a schema, so we need to
    # compare the field & template signature rather than just rely on ID. If
    # the schemas don't match, we increment the mid and try again, creating a
    # new model if necessary.

    def _prepareModels(self):
        "Prepare index of schema hashes."
        self._modelMap = {}

    def _mid(self, srcMid):
        "Return local id for remote MID."
        # already processed this mid?
        if srcMid in self._modelMap:
            return self._modelMap[srcMid]
        mid = srcMid
        srcModel = self.src.models.get(srcMid)
        srcScm = self.src.models.scmhash(srcModel)
        while True:
            # missing from target col?
            if not self.dst.models.have(mid):
                # copy it over
                model = srcModel.copy()
                model['id'] = mid
                model['mod'] = intTime()
                model['usn'] = self.col.usn()
                self.dst.models.update(model)
                break
            # there's an existing model; do the schemas match?
            dstModel = self.dst.models.get(mid)
            dstScm = self.dst.models.scmhash(dstModel)
            if srcScm == dstScm:
                # they do; we can reuse this mid
                model = srcModel.copy()
                model['id'] = mid
                model['mod'] = intTime()
                model['usn'] = self.col.usn()
                self.dst.models.update(model)
                break
            # as they don't match, try next id
            mid += 1
        # save map and return new mid
        self._modelMap[srcMid] = mid
        return mid

    # Decks
    ######################################################################

    def _did(self, did):
        "Given did in src col, return local id."
        # already converted?
        if did in self._decks:
            return self._decks[did]
        # get the name in src
        g = self.src.decks.get(did)
        name = g['name']
        # if there's a prefix, replace the top level deck
        if self.deckPrefix:
            tmpname = "::".join(name.split("::")[1:])
            name = self.deckPrefix
            if tmpname:
                name += "::" + tmpname
        # manually create any parents so we can pull in descriptions
        head = ""
        for parent in name.split("::")[:-1]:
            if head:
                head += "::"
            head += parent
            idInSrc = self.src.decks.id(head)
            self._did(idInSrc)
        # create in local
        newid = self.dst.decks.id(name)
        # pull conf over
        if 'conf' in g and g['conf'] != 1:
            conf = self.src.decks.getConf(g['conf'])
            self.dst.decks.save(conf)
            self.dst.decks.updateConf(conf)
            g2 = self.dst.decks.get(newid)
            g2['conf'] = g['conf']
            self.dst.decks.save(g2)
        # save desc
        deck = self.dst.decks.get(newid)
        deck['desc'] = g['desc']
        self.dst.decks.save(deck)
        # add to deck map and return
        self._decks[did] = newid
        return newid

    # Cards
    ######################################################################

    def _importCards(self):
        # build map of (guid, ord) -> cid and used id cache
        self._cards = {}
        existing = {}
        for guid, ord, cid in self.dst.db.execute(
            "select f.guid, c.ord, c.id from cards c, notes f "
            "where c.nid = f.id"):
            existing[cid] = True
            self._cards[(guid, ord)] = cid
        # loop through src
        cards = []
        revlog = []
        cnt = 0
        usn = self.dst.usn()
        aheadBy = self.src.sched.today - self.dst.sched.today
        for card in self.src.db.execute(
            "select f.guid, f.mid, c.* from cards c, notes f "
            "where c.nid = f.id"):
            guid = card[0]
            if guid in self._changedGuids:
                guid = self._changedGuids[guid]
            if guid in self._ignoredGuids:
                continue
            # does the card's note exist in dst col?
            if guid not in self._notes:
                continue
            dnid = self._notes[guid]
            # does the card already exist in the dst col?
            ord = card[5]
            if (guid, ord) in self._cards:
                # fixme: in future, could update if newer mod time
                continue
            # doesn't exist. strip off note info, and save src id for later
            card = list(card[2:])
            scid = card[0]
            # ensure the card id is unique
            while card[0] in existing:
                card[0] += 999
            existing[card[0]] = True
            # update cid, nid, etc
            card[1] = self._notes[guid][0]
            card[2] = self._did(card[2])
            card[4] = intTime()
            card[5] = usn
            # review cards have a due date relative to collection
            if card[7] in (2, 3) or card[6] == 2:
                card[8] -= aheadBy
            # if odid true, convert card from filtered to normal
            if card[15]:
                # odid
                card[15] = 0
                # odue
                card[8] = card[14]
                card[14] = 0
                # queue
                if card[6] == 1: # type
                    card[7] = 0
                else:
                    card[7] = card[6]
                # type
                if card[6] == 1:
                    card[6] = 0
            cards.append(card)
            # we need to import revlog, rewriting card ids and bumping usn
            for rev in self.src.db.execute(
                "select * from revlog where cid = ?", scid):
                rev = list(rev)
                rev[1] = card[0]
                rev[2] = self.dst.usn()
                revlog.append(rev)
            cnt += 1
        # apply
        self.dst.db.executemany("""
insert or ignore into cards values (?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?)""", cards)
        self.dst.db.executemany("""
insert or ignore into revlog values (?,?,?,?,?,?,?,?,?)""", revlog)
        self.log.append(ngettext("%d card imported.", "%d cards imported.", cnt) % cnt)

    # Media
    ######################################################################

    # note: this func only applies to imports of .anki2. for .apkg files, the
    # apkg importer does the copying
    def _importStaticMedia(self):
        # Import any '_foo' prefixed media files regardless of whether
        # they're used on notes or not
        dir = self.src.media.dir()
        if not os.path.exists(dir):
            return
        for fname in os.listdir(dir):
            if fname.startswith("_") and not self.dst.media.have(fname):
                self._writeDstMedia(fname, self._srcMediaData(fname))

    def _mediaData(self, fname, dir=None):
        if not dir:
            dir = self.src.media.dir()
        path = os.path.join(dir, fname)
        try:
            return open(path, "rb").read()
        except (IOError, OSError):
            return

    def _srcMediaData(self, fname):
        "Data for FNAME in src collection."
        return self._mediaData(fname, self.src.media.dir())

    def _dstMediaData(self, fname):
        "Data for FNAME in dst collection."
        return self._mediaData(fname, self.dst.media.dir())

    def _writeDstMedia(self, fname, data):
        path = os.path.join(self.dst.media.dir(),
                            unicodedata.normalize("NFC", fname))
        try:
            open(path, "wb").write(data)
        except (OSError, IOError):
            # the user likely used subdirectories
            pass

    def _mungeMedia(self, mid, fields):
        fields = splitFields(fields)
        def repl(match):
            fname = match.group("fname")
            srcData = self._srcMediaData(fname)
            dstData = self._dstMediaData(fname)
            if not srcData:
                # file was not in source, ignore
                return match.group(0)
            # if model-local file exists from a previous import, use that
            name, ext = os.path.splitext(fname)
            lname = "%s_%s%s" % (name, mid, ext)
            if self.dst.media.have(lname):
                return match.group(0).replace(fname, lname)
            # if missing or the same, pass unmodified
            elif not dstData or srcData == dstData:
                # need to copy?
                if not dstData:
                    self._writeDstMedia(fname, srcData)
                return match.group(0)
            # exists but does not match, so we need to dedupe
            self._writeDstMedia(lname, srcData)
            return match.group(0).replace(fname, lname)
        for i in range(len(fields)):
            fields[i] = self.dst.media.transformNames(fields[i], repl)
        return joinFields(fields)

    # Post-import cleanup
    ######################################################################

    def _postImport(self):
        for did in self._decks.values():
            self.col.sched.maybeRandomizeDeck(did)
        # make sure new position is correct
        self.dst.conf['nextPos'] = self.dst.db.scalar(
            "select max(due)+1 from cards where type = 0") or 0
        self.dst.save()



# -*- coding: utf-8 -*-
# Copyright: Damien Elmes <anki@ichi2.net>
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

import zipfile, os
import unicodedata
from anki.utils import tmpfile, json
from anki.importing.anki2 import Anki2Importer

class AnkiPackageImporter(Anki2Importer):

    def run(self):
        # extract the deck from the zip file
        self.zip = z = zipfile.ZipFile(self.file)
        col = z.read("collection.anki2")
        colpath = tmpfile(suffix=".anki2")
        open(colpath, "wb").write(col)
        self.file = colpath
        # we need the media dict in advance, and we'll need a map of fname ->
        # number to use during the import
        self.nameToNum = {}
        for k, v in json.loads(z.read("media")).items():
            self.nameToNum[v] = k
        # run anki2 importer
        Anki2Importer.run(self)
        # import static media
        for file, c in self.nameToNum.items():
            if not file.startswith("_") and not file.startswith("latex-"):
                continue
            path = os.path.join(self.col.media.dir(),
                                unicodedata.normalize("NFC", file))
            if not os.path.exists(path):
                open(path, "wb").write(z.read(c))

    def _srcMediaData(self, fname):
        if fname in self.nameToNum:
            return self.zip.read(self.nameToNum[fname])
        return None



from setuptools import setup

setup(
    name = 'vexer',
    version = '0.2',
    description = 'Vocabulary expander using anki and mac osx dictionary.',
    url = 'https://github.com/jlitven/vexer',
    download_url = 'https://github.com/jlitven/vexer/releases/download/v0.1/vexer-0.1.tar.gz',
    keywords = ['anki', 'english', 'vocabulary'],
    author = 'Joshua Litven',
    author_email = 'jlitven@gmail.com',
    license = 'MIT',
    packages = ['vexer', 'anki', 'vexer.card_creator', 'anki.importing',
    'anki.template'],
    package_dir = {'vexer': 'src/vexer',
                    'anki': 'src/anki',
                    'vexer.card_creator': 'src/vexer/card_creator',
                    'anki.importing': 'src/anki/importing',
                    'anki.template': 'src/anki/template'},
    entry_points={
        'console_scripts': [
            'vexer = vexer.vocabulary_creator:main',
        ],
    },
)



# -*- coding: utf-8 -*-
# Copyright: Damien Elmes <anki@ichi2.net>
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

import os
import time

try:
    from pysqlite2 import dbapi2 as sqlite
    vi = sqlite.version_info
    if vi[0] > 2 or vi[1] > 6:
        # latest pysqlite breaks anki
        raise ImportError()
except ImportError:
    from sqlite3 import dbapi2 as sqlite

Error = sqlite.Error

class DB(object):
    def __init__(self, path, text=None, timeout=0):
        encpath = path
        if isinstance(encpath, unicode):
            encpath = path.encode("utf-8")
        self._db = sqlite.connect(encpath, timeout=timeout)
        if text:
            self._db.text_factory = text
        self._path = path
        self.echo = os.environ.get("DBECHO")
        self.mod = False

    def execute(self, sql, *a, **ka):
        s = sql.strip().lower()
        # mark modified?
        for stmt in "insert", "update", "delete":
            if s.startswith(stmt):
                self.mod = True
        t = time.time()
        if ka:
            # execute("...where id = :id", id=5)
            res = self._db.execute(sql, ka)
        else:
            # execute("...where id = ?", 5)
            res = self._db.execute(sql, a)
        if self.echo:
            #print a, ka
            print sql, "%0.3fms" % ((time.time() - t)*1000)
            if self.echo == "2":
                print a, ka
        return res

    def executemany(self, sql, l):
        self.mod = True
        t = time.time()
        self._db.executemany(sql, l)
        if self.echo:
            print sql, "%0.3fms" % ((time.time() - t)*1000)
            if self.echo == "2":
                print l

    def commit(self):
        t = time.time()
        self._db.commit()
        if self.echo:
            print "commit %0.3fms" % ((time.time() - t)*1000)

    def executescript(self, sql):
        self.mod = True
        if self.echo:
            print sql
        self._db.executescript(sql)

    def rollback(self):
        self._db.rollback()

    def scalar(self, *a, **kw):
        res = self.execute(*a, **kw).fetchone()
        if res:
            return res[0]
        return None

    def all(self, *a, **kw):
        return self.execute(*a, **kw).fetchall()

    def first(self, *a, **kw):
        c = self.execute(*a, **kw)
        res = c.fetchone()
        c.close()
        return res

    def list(self, *a, **kw):
        return [x[0] for x in self.execute(*a, **kw)]

    def close(self):
        self._db.close()

    def set_progress_handler(self, *args):
        self._db.set_progress_handler(*args)

    def __enter__(self):
        self._db.execute("begin")
        return self

    def __exit__(self, exc_type, *args):
        self._db.close()

    def totalChanges(self):
        return self._db.total_changes

    def interrupt(self):
        self._db.interrupt()



#! /usr/bin/python

'''
Searches for a word and prints a nicely formatted definition.
'''
import sys, re, pdb
# needed for mac osx dictionary
sys.path.append('/System/Library/Frameworks/Python.framework/Versions/2.7/Extras/lib/python/PyObjC')
from DictionaryServices import DCSCopyTextDefinition
from dictionary_wrappers import DictionaryEntry, Definition

def add_period(string):
    if not string[-1] == '.':
        return string + '.'
    else:
        return string

def first_word(string):
    return string.split(' ', 1)[0]

def create_dictionary_entry(result,
                            num_definitions,
                            num_parts_of_speech):
    '''
    Create a dictionary entry from a dictionary lookup.
    '''
    word = first_word(result)

    # Remove all pronounciation
    pronounciation_regex = r'|*?|'

    symbol = u'\u25b6'
    capital_words = ['PHRASES', 'DERIVATIVES', 'ORIGIN']

    # Create separator regex
    separator = '(' + '|'.join(capital_words + [symbol]) + ')'
    regex = separator + r'(.*?)' + separator
    separated_text = re.split(regex, result)

    # Get parts of speech text
    parts_of_speech_text = []
    if len(separated_text) == 1:
        # no split occured
        symbol_index = result.find(symbol)
        parts_of_speech_text.append(result[symbol_index+1:])
    else:
        for index, text in enumerate(separated_text):
            if symbol in text:
                parts_of_speech_text.append(separated_text[index+1])

    dict_entry = DictionaryEntry(word)
    for text in parts_of_speech_text:

        part_of_speech = first_word(text)

        # Get rid of () pattern following the part of speech
        bracket_regex = part_of_speech + r' \(.*?\)'
        text = re.sub(bracket_regex, '', text, count=1)

        # Split the text by numbers, otherwise by first word
        split_text = re.split(r' \d ', text)
        if len(split_text) == 1:
            split_text = text.split(' ', 1)

        for definitions in split_text[1:]:

            # Only grab first definition
            dot = u'\u2022'
            dot_index = definitions.find(dot)
            if dot_index > 0:
                definition_text = definitions[:dot_index]
            else:
                definition_text = definitions
            colon_index = definition_text.find(':')
            if colon_index > 0:
                description = definition_text[:colon_index].strip()
            else:
                description = definition_text
            description = add_period(description)

            usages = []
            if colon_index > 0:
                usages = definition_text[colon_index+1:].split('|')
                usages = [add_period(usage.strip())
                          for usage in usages]

            definition = Definition(description, usages)
            dict_entry.add_definition(part_of_speech, definition)
            if len(dict_entry[part_of_speech]) == num_definitions:
                break
        if len(dict_entry.items()) == num_parts_of_speech:
            break
    return dict_entry

def raw_entry(word):
    word_range = (0, len(word))
    dictionary_result = DCSCopyTextDefinition(None, word, word_range)
    if not dictionary_result:
        print "{} not found in Dictionary.".format(word)
        return None
    else:
        return dictionary_result

def dictionary_entry(word, num_definitions=1, num_parts_of_speech=3):
    '''Returns a dictonary entry of the word'''
    dictionary_lookup = raw_entry(word)
    if dictionary_lookup:
        return create_dictionary_entry(dictionary_lookup,
                                       num_definitions,
                                       num_parts_of_speech)

def main():
    try:
        word = sys.argv[1]
    except IndexError:
        print 'You did not enter a word to look up.'
        sys.exit()

    print dictionary_entry(word)

if __name__ == '__main__':
    main()
